// trackingchanges.cpp : Defines the entry point for the application.
//

#include "stdafx.h"


// struct change : model that describes changes between string revisions

typedef struct _change {

	typedef enum enumType { addition, deletion };

	long offset, length;
	enumType what;
	bool which;
} change;




// struct node : a simple chained-list implementation

struct node 
{
	change* elem;
	node* next;

	node(change* c) : elem(c), next(NULL)
	{}
	~node()
	{
		delete next; // chain the deletion
		delete elem; // this object
	}

	node* newNode(change* c)
	{
		node* n = new node(c);
		next = n;

		return n;
	}

	node* newNode(long offset, long length, change::enumType what, bool which)
	{
		if (length == 0) return NULL;

		node* n = this;
		while (n->next)
			n = n->next;

		change* c = new change();
		if (!c) 
			return NULL;

		c->offset = offset;
		c->length = length;
		c->what = what;
		c->which = which;

		n->next = new node(c);
		return n->next;
	}

	void dump()
	{
		OutputDebugString("\r\n\r\ndumping of all nodes\r\n--------------------\r\n\r\n");

		node* n = this;
		char s[64];

		do
		{
			if (n->elem)
			{
				sprintf(s,"[%d,%d] what=%s which=%d\r\n", 
					n->elem->offset,
					n->elem->offset + n->elem->length - 1,
					n->elem->what == change::addition ? "addition" : ( n->elem->what == change::deletion ? "deletion" : "change" ),
					n->elem->which ? 2 : 1);
				OutputDebugString(s);
			}

			n = n->next;
		}
		while (n);

		OutputDebugString("\r\n");
	}

};




#define SUBSTRINGLEN 3

// CTrackChanges : a simple change tracking implementation

class CTrackChanges
{
	public:

		// function : findSubstring
		// purpose : find substring s in s2
		char* findSubstring(char* s, int s_len, char* s2)
		{
			if (!s || !s2 || s_len <= 0) return NULL;

			// make sure that the substring fits the definition
			int n_len = 0;
			char* s_tmp = s;
			while (*s_tmp && n_len < s_len)
			{
				n_len++;
				s_tmp++;
			}

			memcpy(_buffer,s,n_len);
			_buffer[n_len] = 0;

			// regular C-runtime finder
			char* result = strstr(s2,_buffer);

			return result;
		}

		// function : evalDiffLen
		// purpose : get to know whether something has been added or removed from s1 ;
		//           also move cursors accordingly
		long evalDiffLen(node* n, 
						 char* s1_begin, char* s2_begin,
						 char* s1, char* s2, 
						 char** s1_new, char** s2_new)
		{
			// is something added in s2 ?
			char* p = findSubstring(s1, SUBSTRINGLEN, s2);
			if (p)
			{
				n->newNode(long(s2 - s2_begin), long(p - s2), change::addition, true); // s2

				// move cursors
				*s1_new = s1;
				*s2_new = p;

				return long(p - s2); // return len of pattern, for debug purpose
			}

			// is something removed in s1 ?
			//
			p = s1;
			char* pold = p;

			p++; // move cursor next

			char* q = NULL;

			// move the cursor of s1 until we find a known substring in s2
			while (*p && (q = findSubstring(p, SUBSTRINGLEN, s2)) == NULL )
				p++;
			
			if (*p == 0)
			{
				// everything left in s1 is a deletion
				// everything left in s2 is an addition
				n->newNode(long(pold - s1_begin), long(p - pold), change::deletion, false); // s1

				// move cursors
				*s1_new = p;
				*s2_new = s2;

				return long(p - pold); // return len of pattern, for debug purpose
			}
			else
			{
				// a fraction of s1 is a deletion
				// a fraction of s2 is an addition
				n->newNode(long(pold - s1_begin), long(p - pold), change::deletion, false); // s1
				n->newNode(long(s2 - s2_begin), long(q - s2), change::addition, true); // s2

				// move cursors
				*s1_new = p;
				*s2_new = q;

				return long(p - pold); // return len of pattern, for debug purpose

			}

			return 0;
		}

		void buildChangeList(node* n, char* s1, char* s2)
		{
			init(n,s1,s2);

			char* p1 = s1;
			char* p2 = s2;

			while( *p1 && *p2 )
			{
				if (*p1 == *p2)
				{
					p1++;
					p2++;
					continue;
				}

				// we have a difference, how long is it?
				char* p1_new;
				char* p2_new;
				long difflen = evalDiffLen(n,s1,s2,p1,p2,&p1_new,&p2_new);
				p1 = p1_new;
				p2 = p2_new;

			} // end while

			if (*p2 == 0)
			{
				// everything left in s1 is a deletion
				if (strlen(p1))
					n->newNode(long(p1 - s1), strlen(p1), change::deletion, false); // s1
			}
			if (*p1 == 0)
			{
				// everything left in s2 is an addition
				if (strlen(p2))
					n->newNode(long(p2 - s2), strlen(p2), change::addition, true); // s2
			}

			clean(n,s1,s2);
		}

		void init(node* n, char* s1, char* s2)
		{
			int slen = max(strlen(s1),strlen(s2));
			_buffer = new char[1 + slen];
		}

		void clean(node* n, char* s1, char* s2)
		{
			delete [] _buffer;
			_buffer = NULL;
		}


		// buildTrackChanges
		// purpose : builds the strings that sums up everything
		//           uses CSS styles to reflect additions/deletions
		char* buildTrackChanges(node* n, char* s1, char* s2)
		{
            const char* szHtmlHeader = "<html>";
			const char* szCSS = "<head><style>\r\n" \
								"body { font-family: Arial; font-size: 0.8em; }\r\n" \
								".tcA { color: #FF0000; }\r\n" \
								".tcD { text-decoration: line-through; font-weight: bold; font-style:italic; }\r\n" \
								"</style></head><body>";
            const char* szHtmlFooter = "</body></html>";

			const char* szCSSAddition_Begin = "<span class=\"tcA\">";
			const char* szCSSAddition_End   = "</span>";
			const char* szCSSDeletion_Begin = "<span class=\"tcD\">";
			const char* szCSSDeletion_End   = "</span>";

			int nHtmlHeader        = strlen(szHtmlHeader);
            int nCSS               = strlen(szCSS);
			int nCSSAddition_Begin = strlen(szCSSAddition_Begin);
			int nCSSAddition_End   = strlen(szCSSAddition_End);
			int nCSSDeletion_Begin = strlen(szCSSDeletion_Begin);
			int nCSSDeletion_End   = strlen(szCSSDeletion_End);
            int nHtmlFooter        = strlen(szHtmlFooter);

			// count the amount of additions and deletions
			int nbAdditions = 0;
			int nbDeletions = 0;
			int nbChanges   = 0;
            node* ntmp = n;
            int len = 0;

			do
			{
				if (ntmp->elem)
				{
                    if (ntmp->elem->what == change::addition)
                        nbAdditions++;
                    else if (ntmp->elem->what == change::deletion)
                        nbDeletions++;

                    len += ntmp->elem->length;
    			}

				ntmp = ntmp->next;
			}
			while (ntmp);

            int n1 = strlen(s1);
            int n2 = strlen(s2);

            // eval the length of the strings when substitions are added to it
            len += nHtmlHeader + nCSS + nHtmlFooter +
                   max(n1,n2) + 
                   nbAdditions * (nCSSAddition_Begin + nCSSAddition_End) +
                   nbDeletions * (nCSSDeletion_Begin + nCSSDeletion_End);

            char* s = new char[1 + len];
            if (!s) return NULL;
            memset(s,0,1+len);

            strcat(s, szHtmlHeader);
            strcat(s, szCSS);

            // now do the substitutions
            //
            int ns1 = 0;
            int ns2 = 0;
            int add1 = 0;
            int add2 = 0;

            ntmp = n;
			do
			{
				if (ntmp->elem)
				{
                    if (ntmp->elem->what == change::addition && 
                        ntmp->elem->which == true)
                    {
                        int ns2_tmp = ntmp->elem->offset;
                        if (ns2_tmp > ns2 && ns2 < n2)
                        {
                            strncat(s, s2 + ns2, ns2_tmp - ns2);
                        }

                        strcat(s, szCSSAddition_Begin);
                        strncat(s, s2 + ns2_tmp, ntmp->elem->length);
                        strcat(s, szCSSAddition_End);
                        ns2 = ntmp->elem->offset + ntmp->elem->length;
                        add2 += ntmp->elem->length;
                        ns1 = ntmp->elem->offset;
                        if (ns1 >= n1)
                            ns1 = n1;
                    }
                    else if (ntmp->elem->what == (int)change::deletion && 
                             ntmp->elem->which == false)
                    {
                        int ns1_tmp = ntmp->elem->offset;
                        if (ns1_tmp > ns1 && ns1 < n1)
                        {
                            strncat(s, s1 + ns1, ns1_tmp - ns1);
                        }

                        strcat(s, szCSSDeletion_Begin);
                        strncat(s, s1 + ns1_tmp, ntmp->elem->length);
                        strcat(s, szCSSDeletion_End);
                        ns1 = ntmp->elem->offset + ntmp->elem->length;
                        ns2 = ntmp->elem->offset + add2;
                        if (ns2 >= n2)
                            ns2 = n2;

                    }

				}

				ntmp = ntmp->next;
			}
			while (ntmp);

            // append the last piece, if any
            if (n2 > ns2)
                strncat(s, s2 + ns2, n2 - ns2);

            strcat(s, szHtmlFooter);

			return s;
		}

	protected:

		char* _buffer;

}; // class CTrackChanges









int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{

	node* n = new node(NULL);
	CTrackChanges c;

    //char* s1 = "hello"; char* s2 = "helldo";
    //char* s1 = "hello"; char* s2 = "helLLdo sir";
    char* s1 = "Two typo here. technical paper"; char* s2 = "Two typosss here. marketing paper";

    c.buildChangeList(n, s1, s2);

	n->dump();

    OutputDebugString("\r\n\r\nstring tracking the changes : \r\n");

    char* s = c.buildTrackChanges(n,s1,s2);

    OutputDebugString(s ? s : "some error occured!");
    OutputDebugString("\r\n");

    delete [] s;

	delete n;

	return 0;
}



