<?php

// trackingChanges
//
// S.Rodriguez - Sept 19 2004 - http://www.arstdesign.com
//
// the purpose of this PHP code is to build a diff between two revisions of a string.
// the resulting diff is aimed to be displayed using html css styles that reflect deletions
// and additions within the strings.
//
// This is a port from a previous C++ code.
//
// How to use it :
//    one line of PHP : echo buildTrackChanges("technical paper", "marketing palkper",1);
//
// will return and output an html string with html headers including the diff :
// <html><head><style>
// body { font-family: Arial; font-size: 0.8em; }
//  .tcA { color: #FF0000; }
//  .tcD { text-decoration: line-through; font-weight: bold; font-style:italic; }
// </style></head>
// <body><span class="tcD">technical</span><span class="tcA">marketing</span> pa<span class="tcA">lk</span>per
// </body></html>
//
//
//  alternatively, use echo buildTrackChanges("technical paper", "marketing palkper",0);
//     to get an html string without the html headers :
//  <span class="tcD">technical</span><span class="tcA">marketing</span> pa<span class="tcA">lk</span>per
//
//
//


// globals
//
$arr_gbl_offset = array();
$arr_gbl_length = array();
$arr_gbl_operation = array();

$SUBSTRINGLEN = 3;

$string1 = "";
$string2 = "";



function findSubstring($s, $s_len, $s2)
{
	global $string1, $string2;

	if ($s == -1 || $s2 == -1 || $s_len <= 0)
		return -1;
	else
	{
		$tmp = substr($string1,$s,$s_len);
		$string2_substr = substr($string2, $s2);
		$index = strpos($string2_substr, $tmp);
		if ($index == false)
			return -1;
		else
			return $s2 + $index;
	}
}



function trackAddition($offset, $length)
{
  global $arr_gbl_offset, $arr_gbl_length, $arr_gbl_operation;

  if ($length == 0) return;
  $n = count($arr_gbl_offset);
  $arr_gbl_offset[$n] = $offset;
  $arr_gbl_length[$n] = $length;
  $arr_gbl_operation[$n] = 1;
}


function trackDeletion($offset, $length)
{
  global $arr_gbl_offset, $arr_gbl_length, $arr_gbl_operation;

  if ($length == 0) return;
  $n = count($arr_gbl_offset);
  $arr_gbl_offset[$n] = $offset;
  $arr_gbl_length[$n] = $length;
  $arr_gbl_operation[$n] = 0;
}



// purpose : get to know whether something has been added or removed from s1 ;
//           also move cursors accordingly
function evalDiffLen($s1, $s2, &$s1_new, &$s2_new)
{
  global $SUBSTRINGLEN, $string1;

  // is something added in s2 ?
  $p = findSubstring($s1, $SUBSTRINGLEN, $s2);
  if ($p != -1)
  {
	trackAddition($s2, $p - $s2);

    // move cursors
    $s1_new = $s1;
    $s2_new = $p;

    return;
  }

  // is something removed from s1 ?
  //
  $p = $s1;
  $pold = $p;

  $p++; // move cursor next

  $q = -1;

  // move the cursor of s1 until we find a known substring in s2
  while (ord($string1{$p}) != 0 && ($q = findSubstring($p, $SUBSTRINGLEN, $s2)) == -1 )
    $p++;
			
  if (ord($string1{$p}) != 0)
  {
	
    // everything left in s1 is a deletion
    // everything left in s2 is an addition
	trackDeletion($pold, $p - $pold);

    // move cursors
    $s1_new = $p;
    $s2_new = $s2;
  }
  else
  {
    // a fraction of s1 is a deletion
    // a fraction of s2 is an addition
	trackDeletion($pold, $p - $pold);
	trackAddition($s2, $q - $s2);
  
    // move cursors
    $s1_new = $p;
    $s2_new = $q;
  }

}

function buildChangeList()
{
  global $string1, $string2;
  global $arr_gbl_offset, $arr_gbl_length, $arr_gbl_operation;

  while( array_pop($arr_gbl_offset) );
  while( array_pop($arr_gbl_length) );
  while( array_pop($arr_gbl_operation) );

  $p1 = 0;
  $p2 = 0;

  while( ord($string1{$p1}) != 0 && ord($string2{$p2}) != 0 )
  {
    if (ord($string1{$p1}) == ord($string2{$p2}))
    {
      $p1++;
      $p2++;
      continue;
    }

    // we have a difference, how long is it?
    $p1_new = -1;
    $p2_new = -1;
	//echo "before evalDiffLen p1=".$p1.",p2=".$p2.",p1_new=".$p1_new.",p2_new=".$p2_new."<br>";
    evalDiffLen($p1,$p2,$p1_new,$p2_new);
	//echo "after evalDiffLen p1=".$p1.",p2=".$p2.",p1_new=".$p1_new.",p2_new=".$p2_new."<br><br>";
    $p1 = $p1_new;
    $p2 = $p2_new;

  } // end while

  if (ord($string2{$p2}) == 0)
  {
    // everything left in s1 is a deletion
	if (ord($string1{$p1}) != 0)
      trackDeletion($p1, strlen(substr($string1, $p1)));
  }
  if (ord($string1{$p1}) == 0)
  {
    // everything left in s2 is an addition
    if (ord($string2{$p2}) != 0)
      trackAddition($p2, strlen(substr($string2, $p2)));
  }

}


// buildTrackChanges
// purpose : builds the strings that sums up everything
//           uses CSS styles to reflect additions/deletions
function buildTrackChanges($oldstring, $newstring, $addCssHeader, $escapeChars)
{
  global $arr_gbl_offset, $arr_gbl_length, $arr_gbl_operation;
  global $string1, $string2;

  $string1 = $oldstring;
  $string2 = $newstring;

  buildChangeList();

  $szHtmlHeader = "<html>";
  $szCSS = "<head><style>\r\n".
           "body { font-family: Arial; font-size: 0.8em; }\r\n".
           ".tcA { color: #FF0000; }\r\n".
           ".tcD { text-decoration: line-through; font-weight: bold; font-style:italic; }\r\n".
           "</style></head><body>";
  $szHtmlFooter = "</body></html>";

  $szCSSAddition_Begin = "<span class=\"tcA\">";
  if ($escapeChars == 1)
	  $szCSSAddition_Begin = "<span class=\\\"tcA\\\">";
  $szCSSAddition_End   = "</span>";
  $szCSSDeletion_Begin = "<span class=\"tcD\">";
  if ($escapeChars == 1)
	$szCSSDeletion_Begin = "<span class=\\\"tcD\\\">";
  $szCSSDeletion_End   = "</span>";

  $nHtmlHeader        = strlen($szHtmlHeader);
  $nCSS               = strlen($szCSS);
  $nCSSAddition_Begin = strlen($szCSSAddition_Begin);
  $nCSSAddition_End   = strlen($szCSSAddition_End);
  $nCSSDeletion_Begin = strlen($szCSSDeletion_Begin);
  $nCSSDeletion_End   = strlen($szCSSDeletion_End);
  $nHtmlFooter        = strlen($szHtmlFooter);

  $n1 = strlen($string1);
  $n2 = strlen($string2);

  $s = "";
  if ($addCssHeader == 1)
  {
    $s .= $szHtmlHeader;
    $s .= $szCSS;
  }

  // now do the substitutions
  //
  $ns1 = 0;
  $ns2 = 0;
  $add1 = 0;
  $add2 = 0;

  // ///////////////
  /* for debug
  for ($i = 0; $i < count($arr_gbl_offset); $i++)
  {
	  echo "o=".$arr_gbl_offset[$i].",l=".$arr_gbl_length[$i].",oper=".$arr_gbl_operation[$i]."<br>";
  }
  echo "<br><br>";
  */
  // ///////////////

  for ($i = 0; $i < count($arr_gbl_offset); $i++)
  {
    $offset = $arr_gbl_offset[$i];
    $length = $arr_gbl_length[$i];
    $operation = $arr_gbl_operation[$i];

	if ($operation == 1) // addition
	{
      $ns2_tmp = $offset;
      if ($ns2_tmp > $ns2 && $ns2 < $n2)
        $s .= substr($string2, $ns2, $ns2_tmp - $ns2);

      $s .= $szCSSAddition_Begin;
      $s .= substr($string2, $ns2_tmp, $length);
      $s .= $szCSSAddition_End;
      $ns2 = $offset + $length;
	  $add2 += $length;
      $ns1 = $offset;
      if ($ns1 >= $n1)
        $ns1 = $n1;
	}
	else // deletion
    {
      $ns1_tmp = $offset;
      if ($ns1_tmp > $ns1 && $ns1 < $n1)
        $s .= substr($string1, $ns1, $ns1_tmp - $ns1);

      $s .= $szCSSDeletion_Begin;
      $s .= substr($string1, $ns1_tmp, $length);
      $s .= $szCSSDeletion_End;
      $ns1 = $offset + $length;
      $ns2 = $offset + $add2;
      if ($ns2 >= $n2)
        $ns2 = $n2;

	}

  } // end for

  // append the last piece, if any
  if ($n2 > $ns2)
    $s .= substr($string2, $ns2, $n2 - $ns2);

  if ($addCssHeader == 1)
    $s .= $szHtmlFooter;

  return $s;

}
/*
echo buildTrackChanges("Two typo here. And another technical paper.<br>", 
"Two typosss he. And another marketing paper.<br>",1);
*/

?>

